from dios.dios import *
from abc import abstractmethod


class _Indexer:
    def __init__(self, _dios):
        self._dios = _dios
        self._data = _dios._data
        self._columns = _dios.columns

    def __setitem__(self, key, val):
        keys, rkey, lowdim = self._unpack_key(key)
        ix, ixalign = self._unpack_rowkey(rkey)
        gen = _unpack_value(keys, ix, val)

        for tup in gen:
            self._set_item(*tup, ixalign=ixalign)

    def __getitem__(self, key):
        keys, rkey, lowdim = self._unpack_key(key)
        ix, ixalign = self._unpack_rowkey(rkey)

        if is_hashable(ix):
            new = pd.Series()
            new.name = ix
        else:
            new = self._dios.copy_empty()

        # set series in new dios OR set values in
        # new series if ix is hashable (see above)
        for k in keys:
            new[k] = self._get_item(k, ix, ixalign=ixalign)
            maby_set_series_name(new[k], k)

        # squeeze to series if a single label was given
        # OR squeeze to val if additional ix is hashable
        if lowdim:
            new = new.squeeze()

        return new

    def _unpack_rowkey(self, rkey):
        align = False
        if is_dios_like(rkey) or is_nested_list_like(rkey):
            raise ValueError("Cannot index with multidimensional key")
        if is_bool_indexer(rkey):
            if not isinstance(rkey, pd.Series):
                raise ValueError("Must pass Series with boolean values only")
            rkey, align = rkey[rkey], True  # kill `False`
        return rkey, align

    @abstractmethod
    def _unpack_key(self, key):
        ...

    @abstractmethod
    def _get_item(self, key, ix, ixalign=False):
        ...

    @abstractmethod
    def _set_item(self, ser, ix, val, ixalign=False):
        ...


class _LocIndexer(_Indexer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # we can use set item here, as this
        # also uses .loc for setting values

    def _set_item(self, *args, **kwargs):
        # we can use DictionaryOfSeries._set_item()
        # here because it also uses .loc
        self._dios._set_item(*args, **kwargs)

    def _get_item(self, key, ix, ixalign=False):
        ser = self._data.loc[key]
        if ixalign:
            ix = ser.index.intersection(ix.index)
        return ser.loc[ix]

    def _unpack_key(self, key):
        lowdim = False
        if isinstance(key, tuple):
            key, ckey, *fail = key
            if fail:
                raise KeyError("To many indexers")
            if is_dios_like(ckey):
                raise ValueError("Cannot index with multidimensional key")
            if is_bool_series(ckey):
                keys = ckey.where(ckey).dropna().index.to_list()
            else:
                if is_hashable(ckey):
                    ckey, lowdim = [ckey], True
                keys = self._data.loc[ckey].index.to_list()
        else:
            keys = self._columns.to_list()
        return keys, key, lowdim


class _iLocIndexer(_Indexer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _set_item(self, key, ix, right, ixalign=False):
        # we do use loc instead of iloc as we get real keys.
        # this works, because keys keep sorted if they come
        # from an index from and series (doesn't work with df)
        ser = self._data[key]
        if ixalign:
            ix = ser.index.intersection(ix.index)
        if isinstance(right, pd.Series):
            left = ser[ix]
            right, ix = align_index_by_policy(left, right)
        ser.iloc[ix] = right

    def _get_item(self, key, ix, ixalign=False):
        ser = self._data.loc[key]
        if ixalign:
            ix = ser.index.intersection(ix.index)
        return ser.iloc[ix]

    def _unpack_key(self, key):
        lowdim = False
        if isinstance(key, tuple):
            key, ckey, *fail = key
            if fail:
                raise KeyError("To many indexers")
            if is_dios_like(ckey):
                raise ValueError("Cannot index with multidimensional key")
            if is_bool_series(ckey):
                keys = ckey.where(ckey).dropna().index.to_list()
            else:
                if is_integer(ckey):
                    ckey, lowdim = [ckey], True
                keys = self._data.iloc[ckey].index.to_list()
        else:
            keys = self._columns.to_list()
        return keys, key, lowdim


def _unpack_value(keys, ix, val):
    """Return a generator that yield (column key, corresponding value, value-align(bool) )
    for all columns.
    This is analogous to DictOfSeries._unpack_value, but with some modifications."""

    # prepare value
    val = list(val) if is_iterator(val) else val
    val = val.squeeze() if is_dios_like(val) else val
    dioslike, nlistlike = is_dios_like(val), is_nested_list_like(val)

    # check value
    if nlistlike and len(val) != len(keys):
        raise ValueError(f"could not broadcast input array with length {len(val)}"
                         f" into dios of length {len(keys)}")
    if dioslike:
        keys = val.columns.intersection(keys)

    for i, k in enumerate(keys):
        if dioslike:
            yield k, ix, val[k]
        elif nlistlike:
            yield k, ix, val[i]
        else:
            yield k, ix, val


def maby_set_series_name(maybe_ser, name):
    if isinstance(maybe_ser, pd.Series):
        maybe_ser.name = name
